/* Pxys - An IRC ProxyScan, m_glineq.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 *
 * $Id: m_glineq.c,v 1.4 2002/04/07 15:10:05 mbuna Exp $
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>             /* close */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "typedefs.h"
#include "conf.h"
#include "misc.h"
#include "debug.h"
#include "m_glineq.h"
#include "m_log.h"
#include "ms_msgout.h"
/* xmalloc at the bottom of include files */
#include "xmalloc.h"


/* Global */
glineq_s *gq = NULL;


void glineq_init()
{
   gq = (glineq_s *) xmalloc(sizeof(glineq_s));
   gq->head = NULL;
   gq->last = NULL;
   gq->size = 0;
   gq->tv.tv_sec = 0;
   gq->tv.tv_usec = 0;
}

void glineq_destroy()
{
   if (gq) {
      glineqItem *p;
      for (p = gq->head; p; p = gq->head) {
         gq->head = gq->head->sv;
         xfree(p);
      }
      xfree(gq);
   }
}

int glineq_push(
   struct in_addr *addr, /* IP address */
   unsigned int type     /* Type of proxy */
)
{
   glineqItem *item;
   
   /* DUPS CHECK */
   for (item = gq->head; item; item = item->sv)
      if (memcmp(&item->addr, addr, sizeof(struct in_addr)) == 0)
         return 0;
   
   item = (glineqItem *) xmalloc(sizeof(glineqItem));
   
   memcpy(&item->addr, addr, sizeof(struct in_addr));
   item->type = type;
   item->sv = NULL;
   
   if (gq->size > 0) {
      gq->last->sv = item;
      gq->last = item;
      gq->size++;
      
      /* Sanity check */
      if (gq->size > 10000) {
         log_error("m_glineq:glineq_push: queue size exceeded 10000!");
         glineq_destroy();
         glineq_init();
         return 0;
      }
      
   } else {
      gq->head = gq->last = item;
      gq->size = 1;
   }
   return 1;
}

glineqItem* glineq_pop()
{
   glineqItem *ret = NULL;
   if (gq->size > 0) {
      ret = gq->head;
      gq->head = gq->head->sv;
      gq->size--;
   }
   return ret;
}

void glineq_popper(struct timeval *tv)
{
   if (gq->size == 0)
      return;
   
   /* Time elapsed since last popper call, and calculate
    * how many glines have to be sent NOW.
    */
   
   if (!gq->tv.tv_sec) {
      gettimeofday(&gq->tv, NULL);
   } else {
      double g_rate, timediff;
      long utimediff;
      glineqItem *gitem;
   
      /* Calc rate */
      g_rate = GLINEQ_FULLDEQUE_DELAY / (double)gq->size;
      
      /* Limit rate by a reasonable value */
      if (g_rate > GLINEQ_MINDEQUE_INTERVAL)
         g_rate = GLINEQ_MINDEQUE_INTERVAL;
      
      /* timer diff */
      utimediff = (tv->tv_sec - gq->tv.tv_sec) * 1000000 + (tv->tv_usec - gq->tv.tv_usec);
      
      timediff = (double) utimediff/1000000.0;
      
      if (timediff >= g_rate) {
         while ((timediff >= g_rate) && (gitem = glineq_pop())) {
            proxy_toast(&gitem->addr, gitem->type, 2 /* cached+queued */);
            xfree(gitem);
            timediff -= g_rate;
         }
         gettimeofday(&gq->tv, NULL);
      }
   }
}


/* m_glineq.c */

  