/* Pxys - An IRC ProxyScan, m_log.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * File logging functions.
 *
 * $Id: m_log.c,v 1.14 2002/04/03 23:51:24 mbuna Exp $
 */



#define TMP_FILE ".pxylog.tmp"


#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif
#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#include "conf.h"
#include "debug.h"
#include "misc.h"

#include "m_log.h"
#include "m_shutdown.h"

/* xmalloc at the bottom of include files */
#include "xmalloc.h"


/* proxy.log
 *
 * Current proxy drop file
 * Format: - TS
 *         - proxies...
 */
 
/* proxy-yesterday.log
 *
 * Format: - TS
 *         - Human TS
 *         - proxies...
 *
 */
 
/* history/
 *
 * proxyYYYYMMDD.log, same format as proxy-yesterday.log
 *
 */

static struct LogFDs {
   int fd;
   char file[256];
} logFDs[8];


pthread_mutex_t mutex_log;
int lastcut_tog = 0;


/* The log_slacken infos */
log_slackinfo_t log_slacken_infos;
#define LOGSLACK_MAXITEMS 32
/* Another separated mutex for dealing with log_slacken */
pthread_mutex_t mutex_log_slacken;


static void
log_open(struct LogFDs *lfd)
{
   if (lfd && lfd->fd < 0)
      lfd->fd = open(lfd->file, O_WRONLY | O_CREAT | O_APPEND, S_IREAD | S_IWRITE);
}

static void
log_stamp()
{
   struct LogFDs* lfd;
   char buf[256];
   time_t now;
   
   time(&now);
   snprintf(buf, sizeof(buf), "%ld\n", now);
   
   lfd = &logFDs[LOGID_PROXY];
   
   log_open(lfd);
   if (lfd->fd < 0) {
      fprintf(stderr, "FATAL: cannot open %s for logging\n", lfd->file);
      log_error("log_stamp: cannot open %s for logging", lfd->file);
      return;
/*      m_shutdown(); */
   }
   
   write(lfd->fd, buf, strlen(buf));
}


/* Init [after pxys's configuration loading] */
void
log_init()
{
   int i;
   
   pthread_mutex_init(&mutex_log, NULL);
   pthread_mutex_init(&mutex_log_slacken, NULL);
   
   snprintf(logFDs[LOGID_PROXY].file, 256, "%s/%s", MyLog->dir, MyLog->current);
   logFDs[LOGID_PROXY].fd = -1;
   
   snprintf(logFDs[LOGID_PXYS].file, 256, "%s/%s", MyLog->dir, MyLog->pxys);
   logFDs[LOGID_PXYS].fd = -1;
   
   snprintf(logFDs[LOGID_GLINES].file, 256, "%s/%s", MyLog->dir, MyLog->glines);
   logFDs[LOGID_GLINES].fd = -1;
   
   snprintf(logFDs[LOGID_PRIVMSG].file, 256, "%s/%s", MyLog->dir, MyLog->privmsg);
   logFDs[LOGID_PRIVMSG].fd = -1;
   
   snprintf(logFDs[LOGID_OPERCMDS].file, 256, "%s/%s", MyLog->dir, MyLog->opercmds);
   logFDs[LOGID_OPERCMDS].fd = -1;
   
   snprintf(logFDs[LOGID_LISTENER].file, 256, "%s/%s", MyLog->dir, MyLog->listener);
   logFDs[LOGID_LISTENER].fd = -1;
   
   snprintf(logFDs[LOGID_STATS].file, 256, "%s/%s", MyLog->dir, MyLog->stats);
   logFDs[LOGID_STATS].fd = -1;
   
   snprintf(logFDs[LOGID_UNREACH].file, 256, "%s/unreach.log", MyLog->dir);
   logFDs[LOGID_UNREACH].fd = -1;
   
   
   /* Open fd's */
   for (i = LOGID_PXYS; i <= LOGID_UNREACH; i++)
      log_open(&logFDs[i]);
}


void
log_write(enum LogIDs id, const char *format, ...)
{
  va_list vl;

  va_start(vl, format);
  log_vwrite(id, format, vl);
  va_end(vl);
}


void
log_vwrite(enum LogIDs id, const char *format, va_list vl)
{
   struct LogFDs* lfd;
   struct iovec vector[3];
   char logbuf[0x400], timebuf[23];
   struct tm ts;
   time_t now;
   
   lfd = &logFDs[id];
   
   pthread_mutex_lock(&mutex_log);
   
   log_open(lfd);
   if (lfd->fd < 0)
      return;
   
   time(&now);
#ifdef HAVE_LOCALTIME_R
   localtime_r(&now, &ts);
#else
   memcpy(&ts, localtime(&now), sizeof(struct tm));
#endif
   
   if (id == LOGID_STATS) {
   /* No datetime in stats log */
      vector[0].iov_len = vsnprintf(logbuf, sizeof(logbuf), format, vl);
      vector[0].iov_base = logbuf;
      vector[1].iov_base = "\n";
      vector[1].iov_len = 1;
      writev(lfd->fd, vector, 2);
   } else {
      vector[0].iov_len = snprintf(timebuf, sizeof(timebuf), "[%d-%d-%d %d:%02d:%02d] ",
                                ts.tm_year + 1900, ts.tm_mon + 1, ts.tm_mday,
   	                            ts.tm_hour, ts.tm_min, ts.tm_sec);
      vector[0].iov_base = timebuf;
      vector[1].iov_len = vsnprintf(logbuf, sizeof(logbuf), format, vl);
      vector[1].iov_base = logbuf;
      vector[2].iov_base = "\n";
      vector[2].iov_len = 1;
      
      writev(lfd->fd, vector, 3);
   }
   
   pthread_mutex_unlock(&mutex_log);
}


void
log_close()
{
   int i;
   
   pthread_mutex_destroy(&mutex_log);
   pthread_mutex_destroy(&mutex_log_slacken);
   
   for (i = LOGID_PROXY; i <= LOGID_STATS; i++)
      if (logFDs[i].fd >= 0)
         close(logFDs[i].fd);
}


/* ROTATING ENGINE FOR LOG FILES
 * 
 * log_rotate_check: Check if we need to sync the log files
 * log_rotate_sync : Check log files TS at startup
 * log_rotate_cut  : Do the cut proxy.log -> proxy-yesterday.log
 * log_rotate_hist : Backup into history directory
 * log_stamp       : Write new TS in current log files
 */
 
 
/* log_rotate_check()
 *  -1 : not rotated
 *  0+ : rotated: returns the number of proxy's in file
 */

int
log_rotate_check(time_t curtime)
{
   int res = -1;
   struct tm ts;
   
#ifdef HAVE_LOCALTIME_R
   localtime_r(&curtime, &ts);
#else
   memcpy(&ts, localtime(&curtime), sizeof(struct tm));
#endif
   
   if(MyLog->update == ts.tm_hour) {
      if(!lastcut_tog) {
         pthread_mutex_lock(&mutex_log);
         close(logFDs[LOGID_PROXY].fd);
         logFDs[LOGID_PROXY].fd = -1;
         res = log_rotate_cut();
         lastcut_tog = 1;
         pthread_mutex_unlock(&mutex_log);
      }
   } else
      lastcut_tog = 0;
   
   return res;
}


/* log_rotate_sync()
 * 
 */

void
log_rotate_sync()
{
   FILE *fp;
   char path[256], buf[256];
   struct tm tm1, tm2;
   char *res;
   time_t last, now;
   
   snprintf(path, sizeof(path), "%s/%s", MyLog->dir, MyLog->current);
   
   if((fp = fopen(path, "r")) == NULL)
      log_stamp();
   else {
      res = fgets(buf, sizeof(buf), fp);
      fclose(fp);
      
      if (res) {
         stripline(buf);
         last = atoi(buf);
         time(&now);
         if(last > 0) {
            if (now - last >= 86400) {
               /* Old proxy.log
                * case 1- it was yesterday: store it in proxy-yesterday.log
                * case 2- it was earlier: store it in history
                */
               
               if(now - last < 172800) /* 2 days */
                  log_rotate_cut();
               else
                  log_rotate_hist(1);
            
            } else if (now - last >= 3600) {
               /* Check if it's already a "current" proxy.log */
#ifdef HAVE_GMTIME_R
               gmtime_r(&last, &tm1); /* Recorded TS */
               gmtime_r(&now, &tm2);  /* NOW! */
#else
               memcpy(&tm1, gmtime(&last), sizeof(struct tm));   /* Recorded TS */
               memcpy(&tm2, gmtime(&now), sizeof(struct tm));    /* NOW */
#endif
            
               if(tm2.tm_hour == tm1.tm_hour) {
                  if(tm2.tm_hour == MyLog->update) {
                     log_rotate_cut();
                  }
               }
/* XXX bugged
                else if(tm2.tm_hour > tm1.tm_hour) {
                  if((tm1.tm_hour < MyLog->update) && (MyLog->update <= tm2.tm_hour)) {
                     log_rotate_cut();
                  }
               } else if(tm2.tm_hour < tm1.tm_hour) {
                  if((tm2.tm_hour >= MyLog->update) || (MyLog->update > tm1.tm_hour)) {
                     log_rotate_cut();
                  }
               }
*/
            } else
               lastcut_tog = 1; /* avoid an autosync by log_rotate_check */
               
         } else {
            unlink(path);
            log_stamp();
         }
      }
   }
}

int
log_rotate_cut()
{
   char path_current[256];
   char path_daily[256];
   char path_tmpfile[256];
   char linebuf[256];
   struct tm tm1, tm2;
   time_t t;
   int nproxy = 0, res;
   FILE *fp, *ftmp;
   
   log_rotate_hist(0);   /* Backup old proxy-yesterday.log */
   
   snprintf(path_current, sizeof(path_current), "%s/%s", MyLog->dir, MyLog->current);
   snprintf(path_daily, sizeof(path_daily), "%s/%s", MyLog->dir, MyLog->daily);
   snprintf(path_tmpfile, sizeof(path_tmpfile), "%s/" TMP_FILE, MyLog->dir);
   
   /* mv proxy.log proxy-yesterday.log */
   res = rename(path_current, path_daily);
   if(res < 0) {
      log_error("log_rotate_cut: rename failed (%s)", strerror(errno));
      
   } else {
      
      /* Converting top TS to humain readable date */
      unlink(path_tmpfile);
      if((ftmp = fopen(path_tmpfile, "a")) != NULL) {
         if((fp = fopen(path_daily, "r")) != NULL) {
            if(fgets(linebuf, sizeof(linebuf), fp)) {
               stripline(linebuf);
               if(*linebuf) {
                  t = atoi(linebuf);
                  
                  fprintf(ftmp, "%s\n", linebuf); /* TS */
                  
#ifdef HAVE_LOCALTIME_R
                  localtime_r(&t, &tm1);
#else
                  memcpy(&tm1, localtime(&t), sizeof(struct tm));
#endif
                  time(&t);
#ifdef HAVE_LOCALTIME_R
                  localtime_r(&t, &tm2);
#else
                  memcpy(&tm2, localtime(&t), sizeof(struct tm));
#endif
                  fprintf(ftmp, "Session started at %d-%d-%d %d:%02d:%02d\n",
                     tm1.tm_year + 1900, tm1.tm_mon + 1, tm1.tm_mday,
                     tm1.tm_hour, tm1.tm_min, tm1.tm_sec);
                  fprintf(ftmp, "Session closed at %d-%d-%d %d:%02d:%02d\n\n",
                     tm2.tm_year + 1900, tm2.tm_mon + 1, tm2.tm_mday,
                     tm2.tm_hour, tm2.tm_min, tm2.tm_sec);
               }
               while(fgets(linebuf, sizeof(linebuf), fp)) {
                  fprintf(ftmp, "%s", linebuf);
                  nproxy++;
               }
            }
            fclose(ftmp);
         }
         fclose(fp);
      }
      res = rename(path_tmpfile, path_daily);
      if(res < 0) {
         pthread_mutex_unlock(&mutex_log);
         log_error("log_rotate_cut: rename (2) failed (%s)\n", strerror(errno));
         pthread_mutex_lock(&mutex_log);
      }
      log_stamp();
   }
   return nproxy;
}


/* resync: 0 -- rotate daily->history
 *         1 -- rotate current->history
 */
void
log_rotate_hist(int resync)
{
   FILE *fp;
   char line[256];
   char pathfrom[256];
   char pathto[256];
   struct tm ts;
   time_t last;
   int res;
   char *str;
   
   snprintf(pathfrom, sizeof(pathfrom), "%s/%s", MyLog->dir,
      resync ? MyLog->current : MyLog->daily);
   
   if((fp = fopen(pathfrom, "r")) == NULL) {
      Debug((DBLMAX, "log_rotate_hist: cannot open file."));
      
      if(resync)
         log_stamp();
      
   } else {
      /* Get recorded TS */
      str = fgets(line, sizeof(line), fp);
      fclose(fp);
      
      if (str) {
         stripline(line);
         if (*line) {
            last = atoi(line);
#ifdef HAVE_GMTIME_R
            gmtime_r(&last, &ts);
#else
            memcpy(&ts, gmtime(&last), sizeof(struct tm));
#endif
            
            snprintf(pathto, sizeof(pathto), "%s/%s/proxy%d-%d-%d.log",
                  MyLog->dir, MyLog->history_dir,
                  ts.tm_year + 1900, ts.tm_mon + 1, ts.tm_mday);
   
            res = rename(pathfrom, pathto);
            if(res < 0) {
               pthread_mutex_unlock(&mutex_log);
               log_error("log_rotate_hist: rename failed (%s)\n", strerror(errno));
               pthread_mutex_lock(&mutex_log);
            }
         }
      }
   }
}


/* Utilities
 *
 */

void
log_error(const char *format, ...)
{
  va_list vl;

  va_start(vl, format);
  log_vwrite(LOGID_PXYS, format, vl);
  va_end(vl);
}


/* Slacken logging
 *
 */

void
log_slacken_init()
{
   /* Allocate an empty array of LOGSLACK_MAXITEMS items for log_slacken */
   log_slacken_infos.size = 0;
   log_slacken_infos.items
        = (log_slackitem_t *)xcalloc(LOGSLACK_MAXITEMS, sizeof(log_slackitem_t));
}

void
log_slacken_destroy()
{
   xfree(log_slacken_infos.items);
}

void
log_slacken(int slack_id, const char *format, ...)
{
   va_list ap;
   
   if (log_slacken_infos.size >= LOGSLACK_MAXITEMS)
      return;
   
   pthread_mutex_lock(&mutex_log_slacken);
   
   log_slacken_infos.items[log_slacken_infos.size].id = slack_id;
   log_slacken_infos.items[log_slacken_infos.size].dumped = 0;
   
   va_start(ap, format);
   vsnprintf(log_slacken_infos.items[log_slacken_infos.size].buf, 256, format, ap);
   va_end(ap);

   log_slacken_infos.size++;
   
   pthread_mutex_unlock(&mutex_log_slacken);
}

void
log_slacken_write()
{
   pthread_mutex_lock(&mutex_log_slacken);
   
   if (log_slacken_infos.size > 0)
      log_slacken_dump();   /* Got something to write */
   
   pthread_mutex_unlock(&mutex_log_slacken);
}

void
log_slacken_dump()
{
   int i, j, cnt;
   
   for (i = 0; i < log_slacken_infos.size; i++) {
      if (!log_slacken_infos.items[i].dumped) {
         for (j = i + 1, cnt = 1; j < log_slacken_infos.size; j++) {
            if (!log_slacken_infos.items[j].dumped
               && (log_slacken_infos.items[i].id == log_slacken_infos.items[j].id)) {
               ++cnt;
               log_slacken_infos.items[j].dumped = 1; /* do not write twice! */
            }
         }
         log_write(LOGID_LISTENER, "%dx %s", cnt, log_slacken_infos.items[i].buf);
      }
   }
   /* Empty! */
   log_slacken_infos.size = 0;
}

