/* Pxys - An IRC ProxyScan, m_opercmds.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * Called by ms_myclient.c
 * 
 *
 * $Id: m_opercmds.c,v 1.42 2002/07/28 23:05:27 mbuna Exp $
 */
 

#define MAX_LASTCOMM 10
#define MAX_LASTLOG 100

#include "config.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "cachehash.h"
#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "match.h"
#include "misc.h"

#include "m_opercmds.h"
#include "ms_msgout.h"
#include "m_array.h"
#include "m_cache.h"
#include "m_clientq.h"
#include "m_glineq.h"
#include "m_log.h"
#include "m_noscan.h"
#include "m_proxycache.h"
#include "m_stats.h"
#include "m_structs.h"
#include "m_shutdown.h"
#include "m_thread.h"
#include "ms_auth.h"
#include "ms_globals.h"
#include "ms_netutils.h"

#include "xmalloc.h"

#ifndef INADDR_NONE
#define INADDR_NONE   0xffffffff
#endif

#ifdef SCAN_WINGATE
#define S_WINGATE "WINGATE "
#else
#define S_WINGATE
#endif

#ifdef SCAN_SOCKS
#define S_SOCKS "SOCKS4/5 "
#else
#define S_SOCKS
#endif

#ifdef SCAN_HTTP
#define S_HTTP "HTTP:"
#else
#define S_HTTP
#endif


extern struct Client * mClient;
extern struct Server * mServer;
extern const char MyClientXXX[];
extern long StartTime;
extern htab *mHS;
extern htab *mHC;
extern htab *pxcacheHash;
extern cachehtab *cachetab;
extern scanstats_s *scanstats;
extern pthread_mutex_t mutex_pxcache;
extern pthread_mutex_t mutex_link;
extern long ipa_current;
extern long totalscanned;
extern long totalfound;
extern time_t startscan_time;
extern int msg_bogus;
extern server_handle_t mSA;
extern long serv_current;
extern long ipa_max, clients_counter;
extern time_t connection_starttime; /* ms_socket.c */
extern glineq_s *gq;
extern int recv_eintr_count,
           send_eintr_count,
           recv_epipe_count,
           send_epipe_count,
           recv_err_count,
           send_err_count;
int unreach_count = 0;
extern long found_in_cache;
extern unsigned long count_malloc, count_calloc, count_free; /* xmalloc */

struct lastcomm {
   time_t tstamp;
   char nick[NICKLEN+1];
   char userip[16];
   char command[32];
   struct lastcomm * next;
};

struct lastlog {
   time_t tstamp;
   char ip[16];
   long toasted;
   char reason[64];
   struct lastlog * next;
};

struct lastcomm * mLastComm = NULL;
int lastcomm_count = 0;

struct lastlog * mLastLog = NULL;
int lastlog_count = 0;

extern statistics_s *pxys_stats;

void
opercmds_help(struct Client *cptr, const char *nn, const char* arg)
{
   char buf[32];
   char *cmd = "help";
   
   if (arg) {
      snprintf(buf, sizeof(buf), "%s %s", cmd, arg);
      lastcomm_put(cptr, buf);
   } else
      lastcomm_put(cptr, cmd);
   
   client_sendto(nn, NETWORK " Proxy Scanner -- pxys %s", VERSION);
   client_sendto(nn, "\2Commands (global opers only):\2");
   client_sendto(nn, "   INV      - Get an invite to the proxy console channel.");
   client_sendto(nn, "   STATUS   - Status command :P~");
   client_sendto(nn, "   STATS    - Hash tables info and scan statistics.");
   client_sendto(nn, "   LINKS    - Show linked servers and noscan status.");
   client_sendto(nn, "   PROXYTOP - Show proxies found per server.");
   client_sendto(nn, "   INFO     - Show client's info.");
   client_sendto(nn, "   LASTLOG  - Last open gates reported.");
   client_sendto(nn, "   GREM     - Remove a gline set by me.");
   client_sendto(nn, "   RECHECK  - Scan <nick> again as soon as possible.");
   
   if (arg && !stricmp(arg, "all")) {
      client_sendto(nn, "   LASTCOMM - Last commands.");
   }
   
   client_sendto(nn, "All commands are logged. Enjoy!");

}

/* Invite request command */
void
opercmds_inv(struct Client *cptr, const char *nn, const char* arg)
{
   lastcomm_put(cptr, "inv");
   
#ifdef N2K_SUPPORT
   sendto_uplink("%s%s " TOK_INVITE " %s %s",
	             MyYY, MyClientXXX, cptr->nick, MyConsole->channel);
#else
   sendto_uplink("%s%s " MSG_INVITE " %s %s",
	             MyYY, MyClientXXX, cptr->nick, MyConsole->channel);
#endif
}


void
opercmds_status(struct Client *cptr, const char *nn, const char* arg)
{
   char buf[256], cpybuf[256];
   char humantime[64];
   time_t now, totalscantime;
#ifdef SCAN_HTTP
   int i;
#endif
   double percent;
   
   lastcomm_put(cptr, "status");
   
   client_sendto(nn, "\2Network status\2");
   
   client_sendto(nn, "Remote clients: \2%ld\2", ipa_current);
   client_sendto(nn, "Remote servers: \2%ld\2", serv_current);
   client_sendto(nn, "Bogus messages: %ld", msg_bogus);
   
   client_sendto(nn, "");
   
   client_sendto(nn, "\2Proxy scanner status\2");

   snprintf(buf, sizeof(buf), "Configuration: " S_WINGATE S_SOCKS S_HTTP);
   if (MyPerform->http_ports_count > 0) {
      int i;
      unsigned short *portp = MyPerform->http_ports;
      for (i = 0; i < MyPerform->http_ports_count; i++) {
         snprintf(cpybuf, sizeof(cpybuf), "%s %u", buf, portp[i]);
         strcpy(buf, cpybuf);
      }
   }
   
   client_sendto(nn, "%s", buf);
   
   percent = 100.0 * (double) scanstats->running_threads / (double) MyPerform->scanthreads;
   client_sendto(nn, "Current activity: \2%g%%\2 - Scan threads:\2%ld\2  RUN:\2%ld\2  WAIT:\2%ld\2",
        percent, MyPerform->scanthreads, scanstats->running_threads, scanstats->waiting_threads);
   
   client_sendto(nn, "Simultaneous scans: \2%ld\2 (Max: %ld)",
        countscans(), SCANS_PER_THREAD * MyPerform->scanthreads);
   
   client_sendto(nn, "Connect() timeout: %ld, send(): %ld, listener: %ld (seconds)",
        MyPerform->timeout_connect,
        MyPerform->timeout_send,
        MyPerform->timeout_listen);
   
   time(&now);
   totalscantime = now - startscan_time;
   if(totalscantime == 0)
      totalscantime = 1;
   
   human_time(totalscantime, humantime);
   percent = (double)totalscanned/(double)totalscantime;
   percent *= 3600.0;
   client_sendto(nn, "\2%ld\2 scans completed. Runtime: %s (average: %.0f/h)",
        totalscanned, humantime, percent);
   
   client_sendto(nn, "Open gates found: \2%ld\2", totalfound);
   
#ifdef SCAN_WINGATE
   if (totalfound > 0)
      percent = 100.0 * (double) pxys_stats->total_wingate
         / (double) totalfound;
   else
      percent = 0.0;
   client_sendto(nn, " - WinGate: \2%ld\2 [%g%%]",
        pxys_stats->total_wingate, percent);
#endif
#ifdef SCAN_SOCKS
   if (totalfound > 0)
      percent = 100.0 * (double) pxys_stats->total_socks4
         / (double) totalfound;
   else
      percent = 0.0;
   client_sendto(nn, " - Socks4: \2%ld\2 [%g%%]",
        pxys_stats->total_socks4, percent);
   if (totalfound > 0)
      percent = 100.0 * (double) pxys_stats->total_socks5
         / (double) totalfound;
   else
      percent = 0.0;
   client_sendto(nn, " - Socks5: \2%ld\2 [%g%%]",
        pxys_stats->total_socks5, percent);
#endif
#ifdef SCAN_HTTP
   if (pxys_stats->total_http != NULL) {
      for (i = 0; pxys_stats->total_http[i].port > 0; i++) {
         if (totalfound > 0)
            percent = 100.0 * (double) pxys_stats->total_http[i].total
               / (double) totalfound;
         else
            percent = 0.0;
         client_sendto(nn, " - HTTP/%u: \2%ld\2 [%g%%]",
              pxys_stats->total_http[i].port, pxys_stats->total_http[i].total, percent);
      }
   }
#endif

   
   client_sendto(nn, "");
   client_sendto(nn, "\2Miscellaneous\2");
   client_sendto(nn, "\37Gline delay\37: %d secs (cached for %ld secs)",
        MyGline->delay, MyGline->cached);
   human_time(now - StartTime, humantime);
   client_sendto(nn, "\37Uptime\37: %s", humantime);
   client_sendto(nn, "\37Version\37: pxys " VERSION " - " __DATE__ " " __TIME__);
}


void
opercmds_links(struct Client *cptr, const char *nn, const char* arg)
{
   struct Server *sptr;
   
   lastcomm_put(cptr, "info");
   
   if (hfirst(mHS)) do {
      sptr = (struct Server *)hstuff(mHS);
      client_sendto(nn, "%4u : %s%s [clients: %u(%u)/%u] (%s%s)", 
      sptr->myx, sptr->bursting ? "\2!\2" : " ",
      sptr->name, sptr->clients, sptr->maxclients, sptr->nn_mask + 1,
      sptr->noscan ? "noscan" : "\2scan\2",
      sptr->shared ? "- \2sharing activated\2" : "");
   } while (hnext(mHS));
}

void
opercmds_info(struct Client *cptr, const char *nn, const char* arg)
{
   char buf[64];
   struct Client *target;
   struct Server *server;
   time_t now, lastcache = 0;
   int noscan = 1;
   
   if (!arg || strlen(arg) > NICKLEN) {
      client_sendto(nn, "\37Usage\37: info <nick>");
      return;
   }
   
   snprintf(buf, sizeof(buf), "info %s", arg);
   lastcomm_put(cptr, buf);
   
   target = array_findclient_from_nick(arg);
   
   if (target) {
      numnick(target, buf);
      client_sendto(nn, "Nick: %s [%s] Numeric: %s",
           target->nick, inet_ntoa(target->ip), buf);
      if (target->flags & FLAG_OPER)
           client_sendto(nn, "%s is a global IRC operator", target->nick);
      
      pthread_mutex_lock(&mutex_link);
      server = mSA[target->nserv];
      if (!server) {
         log_error("opercmds_info: Server not found for %s %u\n", target->nick, target->nserv);
         client_sendto(nn, "ERROR: Server #%u not found!", target->nserv);
         return;
      }
      client_sendto(nn, "%s is on IRC via server %s", target->nick, server->name);
      
      if (server->noscan)
         client_sendto(nn, "%s will not be scanned (using a NOSCAN server)", target->nick);
      else if (server->shared)
         client_sendto(nn, "%s will be scanned by another proxyscanner (shared activated)",
              target->nick);
      else {
         check_noscan_ip(&target->ip, &noscan);
         if (noscan)
            client_sendto(nn, "%s's IP is protected by \2NOSCAN\2", target->nick);
         else {
            if (target->last == 0)
               client_sendto(nn, "%s has not been checked yet", target->nick);
            else {
               time(&now);
               human_time(now - target->last, buf);
               client_sendto(nn, "%s has been checked %s ago", target->nick, buf);
            }
         }
      }
      pthread_mutex_unlock(&mutex_link);
      
      if (!noscan) {
         lastcache = cache_fetch(&target->ip);
         if(lastcache > 0) {
            time(&now);
            human_time(now - lastcache, buf);
            client_sendto(nn, "IP %s has been checked %s ago", inet_ntoa(target->ip), buf);
         }
      }
      
      client_sendto(nn, "End of client info.");
      
   } else
      client_sendto(nn, "\37INFO\37: Unable to find nick \2%s\2.", arg);
}

void
opercmds_lastcomm(struct Client *cptr, const char *nn, const char* arg)
{
   lastcomm_put(cptr, "lastcomm");
   
   if (!arg)
      lastcomm_get(nn, 5);
   else {
      int hist = atoi(arg);
	  lastcomm_get(nn, (hist > 0) ? hist : 5);
   }
}

void
opercmds_lastlog(struct Client *cptr, const char *nn, const char* arg)
{
   lastcomm_put(cptr, "lastlog");
   if (!arg)
      lastlog_get(nn, 5);
   else {
      int hist = atoi(arg);
      lastlog_get(nn, (hist > 0) ? hist : 5);
   }
}

void
opercmds_grem(struct Client *cptr, const char *nn, const char* arg)
{
   const char cmd[] = "grem";
   char buf[32];
   unsigned long ip;
   
   if (arg) {
      snprintf(buf, sizeof(buf), "%s %s", cmd, arg);
      lastcomm_put(cptr, buf);
   } else {
      lastcomm_put(cptr, cmd);
      client_sendto(nn, "\37Usage\37: grem <IP>");
      return;
   }
   
   if ((arg[0] == '*') && (arg[1] == '@'))
      arg += 2;
   
   if (!*arg) /* "*@" ?? */
      return;
   
   ip = inet_addr(arg);
   if(INADDR_NONE == ip) {
      client_sendto(nn, "\37Usage\37: grem <IP>");
   } else {
      struct in_addr addr;
      addr.s_addr = ip;
      
      strncpy(buf, inet_ntoa(addr), sizeof(buf) - 1);
      buf[sizeof(buf) - 1] = '\0';
      
      cache_remove(&addr); /* Remove it from the main cache: it'll allow us to rescan it */
      
      if (proxycache_remove(&addr)) {
         
         log_write(LOGID_GLINES, "-*@%s", buf);
         
#ifdef N2K_SUPPORT
         sendto_uplink("%s " TOK_GLINE " * -%s", MyYY, buf);
#else
         sendto_uplink("%s " MSG_GLINE " * -%s", MyYY, buf);
#endif
         client_sendto(nn, "\2GREM\2: Gline on %s removed.", buf);
      } else
         client_sendto(nn, "\2GREM\2: IP %s not found in my proxy cache", buf);
   }
}


void
opercmds_recheck(struct Client *cptr, const char *nn, const char* arg)
{
   const char cmd[] = "recheck";
   char buf[32];
   struct Client *target;
   struct Server *server;
   int noscan;
   yxx cliyxx;
   
   if (!arg || strlen(arg) > NICKLEN) {
      client_sendto(nn, "\37Usage\37: recheck <nick>");
      return;
   }
   
   snprintf(buf, sizeof(buf), "%s %s", cmd, arg);
   lastcomm_put(cptr, buf);
   
   target = array_findclient_from_nick(arg);
   if (!target) {
      client_sendto(nn, "\37RECHECK\37: Nick %s not found.", arg);
      return;
   }
   
   server = mSA[target->nserv];
   if (!server) {
      log_error("opercmds_recheck: Server not found for %s %u\n", target->nick, target->nserv);
      client_sendto(nn, "ERROR: Server #%u not found!", target->nserv);
      return;
   }
   
   if (server->noscan) {
      client_sendto(nn, "Can't recheck. %s will not be scanned (using a NOSCAN server)",
           target->nick);
      return;
   } else if (server->shared) {
      client_sendto(nn, "Can't recheck. %s will be scanned by another proxyscanner (shared activated)",
           target->nick);
      return;
   }
   
   check_noscan_ip(&target->ip, &noscan);
   if (noscan) {
      client_sendto(nn, "Can't recheck. %s's IP is protected by \2NOSCAN\2", target->nick);
      return;
   }
   
   pthread_mutex_lock(&mutex_link);
   cache_remove(&target->ip);
   proxycache_remove(&target->ip);
   target->last = 0;
   
   /* Queue if sync */
   cliyxx.server_n = target->nserv;
   cliyxx.client_n = target->nnick;
   cliyxx.bogus = 0;
#ifdef N2K_SUPPORT
   cliyxx.extended = 1;
#else
   cliyxx.extended = 0;
#endif
   clientq_queue(&cliyxx);
   pthread_mutex_unlock(&mutex_link);
   
   client_sendto(nn, "\37RECHECK\37: User %s [%s] will be recheck'd for open proxy as soon as possible.",
        target->nick, inet_ntoa(target->ip));
}

void
opercmds_stats(struct Client *cptr, const char *nn, const char* arg)
{
   ub4 buckets, items;
   int qtotal, qsize;
   double dble;
   time_t now;
   /*
#if defined(HAVE_MALLINFO) && !defined(DMALLOC)
   struct mallinfo mi;
#endif
   */
   
   lastcomm_put(cptr, "stats");
   
   hstat(mHC, &buckets, &items, &dble);
   client_sendto(nn, "Clients hash table: %d buckets, %d items, existing: %g",
        buckets, items, dble);
   
   hstat(mHS, &buckets, &items, &dble);
   client_sendto(nn, "Servers hash table: %d buckets, %d items, existing: %g",
        buckets, items, dble);
   
   time(&now);
   dble = (double) clients_counter / (double) (now - connection_starttime);
   client_sendto(nn, "Max clients: %ld - Clients counter: %ld - Average: %g/s",
        ipa_max, clients_counter, dble);
        
   clientq_stats(&qtotal, &qsize);
   client_sendto(nn, "ClientQ size: %d total, %d used", qtotal, qsize);
   
   cache_hstat(cachetab, &buckets, &items);
   if (MyPerform->cachesize)
      dble = 100.0 * (double)items / (double)MyPerform->cachesize;
   else
      dble = 0.0;
   client_sendto(nn, "IP cache: %d total, %d buckets, %d items (%f%% used)",
        MyPerform->cachesize, buckets, items, dble);
   
   pthread_mutex_lock(&mutex_pxcache);
   client_sendto(nn, "Proxy cache: %d items (cached for %d secs)",
        hcount(pxcacheHash), MyGline->cached);
   pthread_mutex_unlock(&mutex_pxcache);
   
   client_sendto(nn, "GlineQ: %d items", gq->size);

   client_sendto(nn, "Found-in-cache counter: %ld", found_in_cache);

   client_sendto(nn, "Hosts unreachable: %d", unreach_count);
   
   client_sendto(nn, "EINTR: recv: %d, send: %d", recv_eintr_count, send_eintr_count);
   client_sendto(nn, "EPIPE: recv: %d, send: %d", recv_epipe_count, send_epipe_count);
   client_sendto(nn, "Other: recv: %d, send: %d", recv_err_count, send_err_count);
   
   client_sendto(nn, "malloc: %u - calloc: %u - free: %u",
                 count_malloc, count_calloc, count_free);

/*
#if defined(HAVE_MALLINFO) && !defined(DMALLOC)  
   mi=mallinfo();
   client_sendto(nn, "\2libc malloc stats:\2");
   client_sendto(nn, "Total space allocated from system: %ld", (long int)mi.arena);
   client_sendto(nn, "Number of non-inuse chunks       : %ld", (long int)mi.ordblks);
   client_sendto(nn, "mmap()'ed regions                : %ld", (long int)mi.hblks);
   client_sendto(nn, "Total space in mmaped regions    : %ld", (long int)mi.hblkhd);
   client_sendto(nn, "Total allocated space            : %ld", (long int)mi.uordblks);
   client_sendto(nn, "Total non-inuse space            : %ld", (long int)mi.fordblks);
   client_sendto(nn, "Top-most, releaseable space      : %ld", (long int)mi.keepcost);
#endif
   */
   if (arg && !stricmp(arg, "all"))
      stats_dump(nn);
}

void
opercmds_wannounce(struct Client *cptr, const char *nn, const char* arg)
{
   char buf[160];
   
   if (!arg) {
      client_sendto(nn, "\37Usage\37: wannounce <wallops message>");
      return;
   }
   
   snprintf(buf, sizeof(buf), "wannounce %s", arg);
   lastcomm_put(cptr, buf);
   
   sendto_uplink("%s WA :%s", MyYY, arg);
}


void
opercmds_showopr(struct Client *cptr, const char *nn, const char* arg)
{
   char buf[160];
   
   if (!arg || strlen(arg) > 3) {
      client_sendto(nn, "\37Usage\37: showopr <on/off>");
      return;
   }
   
   snprintf(buf, sizeof(buf), "showopers %s", arg);
   lastcomm_put(cptr, buf);
   
   if (!stricmp(arg, "0") || !stricmp(arg, "off")) {
      if (cptr->flags & FLAG_SHOWOPR)
         cptr->flags &= ~FLAG_SHOWOPR;
      
      client_sendto(nn, "SHOWOPR disabled.");
      
   } else if (!stricmp(arg, "1") || !stricmp(arg, "on")) {
      if (!(cptr->flags & FLAG_SHOWOPR))
         cptr->flags |= FLAG_SHOWOPR;
      
      client_sendto(nn, "SHOWOPR activated.");
   } else
      client_sendto(nn, "\37Usage\37: showopr <on/off>");
}

void
opercmds_listshowopr(struct Client *cptr, const char *nn, const char* arg)
{
   int i = 0;
   struct Client *client;
   char *servername;
   
   lastcomm_put(cptr, "listshowopr");
   
   client_sendto(nn, "List of clients using SHOWOPR:");
	         
   if (hfirst(mHC)) do
   {
       client = (struct Client *) hstuff(mHC);
	   if (client->flags & FLAG_SHOWOPR) {
	      if (mSA[client->nserv])
	         servername = ((struct Server *)mSA[client->nserv])->name;
	      else
	         servername = "*DESYNC*";
	      
	      client_sendto(nn, "   %9s%s (%s)", client->nick,
	         (client->flags & FLAG_OPER) ? "*" : " ", servername);
	      
	      ++i;
	   }
	   
   } while(hnext(mHC));
   
   client_sendto(nn, "End of SHOWOPR list (%d found).", i);
}

void
opercmds_dropq(struct Client *cptr, const char *nn, const char* arg)
{
   int qdummy, qsize;
   struct Client *client;

   lastcomm_put(cptr, "dropq");

   clientq_stats(&qdummy, &qsize);
   client_sendto(nn, "DROPQ: %d users in clientQ", qsize);

   clientq_drop(qsize);

   client = array_findclient_from_numeric(nn);
   if (client) {
      client_sendto_console("DROPQ by %s (%d IPs)", client->nick, qsize);
   } else {
      client_sendto_console("DROPQ: desync! (numnick: %s)", nn);
   }
   
   clientq_stats(&qdummy, &qsize);
   client_sendto(nn, "DROPQ: done. (%d users in clientQ)", qsize);
}

static int
proxytop_comparator(const void *s1, const void *s2)
{
   struct Server *server1 = *(struct Server **)s1;
   struct Server *server2 = *(struct Server **)s2;
   unsigned int t1 = server1->currtoasted + server1->lasttoasted;
   unsigned int t2 = server2->currtoasted + server2->lasttoasted;
   int result;

   if (t1 < t2)
      result = 1;
   else if (t1 > t2)
      result = -1;
   else
      result = strcmp(server1->name, server2->name);

   return result;
}

void
opercmds_proxytop(struct Client *cptr, const char *nn, const char* arg)
{
   unsigned int total = 0;
   int i;
   
   struct Server **sBase = (struct Server **) xmalloc(hcount(mHS)
                                                    * sizeof(struct Server *));
   
   struct Server **p = sBase;
   
   lastcomm_put(cptr, "proxytop");
   
   if (hfirst(mHS)) do {
      *p++ = hstuff(mHS);
   } while (hnext(mHS));
   
   qsort(sBase, hcount(mHS), sizeof(struct Server *), proxytop_comparator);
   
   for (i = 0, p = sBase; i < hcount(mHS); i++, p++)
      total += (*p)->currtoasted + (*p)->lasttoasted;
   
   if (total == 0)
      total = 1;
   
   for (i = 0, p = sBase; i < hcount(mHS); i++, p++)
      client_sendto(nn, "%d %s%s \2%u\2 (%u) - %g%%",
         i + 1,
         (*p)->bursting ? "\2!\2" : " ",
         (*p)->name,
         (*p)->currtoasted + (*p)->lasttoasted,
         (*p)->currtoasted,
         100.0 * (double)((*p)->currtoasted + (*p)->lasttoasted)/(double)total);
   
   xfree(sBase);
}


void
lastcomm_put(const struct Client *cptr, const char *cmd)
{
   struct lastcomm *last, *ptr;
   
   if(lastcomm_count < MAX_LASTCOMM) {
      last = (struct lastcomm *) xmalloc(sizeof(struct lastcomm));
      memset(last, 0, sizeof(struct lastcomm));
      
      if(NULL == mLastComm)
         mLastComm = last;
      else {
         ptr = mLastComm;
         while(ptr->next) { ptr = ptr->next; }
         ptr->next = last;
      }
      
      ++lastcomm_count;
   
   } else {
      last = mLastComm;
      mLastComm = mLastComm->next;
      for(ptr = mLastComm; ptr && ptr->next; ptr = ptr->next)
         ;
      ptr->next = last;
   }
   time(&last->tstamp);
   strncpy(last->nick, cptr->nick, NICKLEN);
   last->nick[NICKLEN] = '\0';
   strncpy(last->userip, inet_ntoa(cptr->ip), 16 - 1);
   strncpy(last->command, cmd, 32 - 1);
   
   last->next = NULL;
      
   log_write(LOGID_OPERCMDS, "%s (%s) : %s", last->nick, last->userip, last->command);
}

void
lastcomm_get(const char *to, int hist)
{
   struct lastcomm *last;
   int i;
   
   if(hist > MAX_LASTCOMM) {
      client_sendto(to, "Max lastcomm history length is %d.", MAX_LASTCOMM);
   } else {
      if(mLastComm == NULL) {
         client_sendto(to, "Nothing in history.");
      } else {
         last = mLastComm;
         if(lastcomm_count > hist) {
            for (i = 0; last && (i < lastcomm_count - hist); last = last->next, i++);
         }
         for(i = 0; last && (i < hist); last = last->next, ++i)
            client_sendto(to, "%s %s [%s] %s", hhmmss(last->tstamp), last->nick, last->userip, last->command);
      }
   }
}

void
lastcomm_free()
{
   struct lastcomm *ptr, *pnext;
   
   for (ptr = mLastComm; ptr; ptr = pnext) {
      pnext = ptr->next;
      xfree(ptr);
   }
   
   mLastComm = NULL;
}

void
lastlog_put(const char *ip, long toasted, const char *reason)
{
   struct lastlog *last, *ptr;
   
   if(lastlog_count < MAX_LASTLOG) {
      last = (struct lastlog *) xmalloc(sizeof(struct lastlog));
      memset(last, 0, sizeof(struct lastlog));
      
      if(NULL == mLastLog)
         mLastLog = last;
      else {
         ptr = mLastLog;
         while (ptr->next)
            ptr = ptr->next;
         ptr->next = last;
      }
      
      ++lastlog_count;
      
   } else {
      last = mLastLog;
      mLastLog = mLastLog->next;
      for(ptr = mLastLog; ptr && ptr->next; ptr = ptr->next)
         ;
      ptr->next = last;
   }
   time(&last->tstamp);
   strncpy(last->ip, ip, 16 - 1);
   last->toasted = toasted;
   strncpy(last->reason, reason, 64 - 1);
   last->next = NULL;
}

void
lastlog_get(const char *to, int hist)
{
   struct lastlog *last;
   int i;
   
   if(hist > MAX_LASTLOG) {
      client_sendto(to, "Max lastlog history length is %d.", MAX_LASTLOG);
   } else {
      if(mLastLog == NULL)
         client_sendto(to, "Nothing in history.");
      else {
         last = mLastLog;
         if(lastlog_count > hist) {
            for (i = 0; last && (i < lastlog_count - hist); last = last->next, i++);
         }
         for(i = 0; last && (i < hist); last = last->next, ++i)
            client_sendto(to, "%s *@%s [%ld] %s", hhmmss(last->tstamp), last->ip, last->toasted, last->reason);
      }
   }
}


void
lastlog_free()
{
   struct lastlog *ptr, *pnext;
   
   for (ptr = mLastLog; ptr; ptr = pnext) {
      pnext = ptr->next;
      xfree(ptr);
   }
   
   mLastLog = NULL;
}


/* TODO: I SUCK */
char *
hhmmss(time_t tstamp)
{
   static char outbuf[32];
   struct tm ts;
   memcpy(&ts, localtime(&tstamp), sizeof(struct tm));
   
   snprintf(outbuf, sizeof(outbuf), "%d-%d-%d %d:%02d:%02d",
        ts.tm_year + 1900, ts.tm_mon + 1, ts.tm_mday,
        ts.tm_hour, ts.tm_min, ts.tm_sec);
   
   return outbuf;
}

/* ms_myclient.c */
