/* Pxys, IRC ProxyScan, m_proxycache.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * 
 * Added in version 1.5, proxycache module handles the caching of the insecure
 * proxies previously found by the proxyscanner.
 * If you checked the sources, you're gonna tell me: heh, but you are _already_
 * storing glines in ms_glines.c, in a chained list. Hey! That's why!
 * We're going to check this cache *every time* an user connect. This will also
 * allow us to set short glines (eg. about 15 min), to avoid too much glines
 * in all ircus on the network. But if a proxy come back after these 15 min,
 * proxycache will have it and will gline it for 15 min _immediatly_.
 *
 * $Id: m_proxycache.c,v 1.16 2002/07/19 13:34:20 mbuna Exp $
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <sys/types.h>
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include "m_structs.h"
#include "typedefs.h"

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "m_cache.h"
#include "m_log.h"
#include "m_proxycache.h"
#include "ms_msgout.h"
/* xmalloc at the bottom of include files */
#include "xmalloc.h"

htab *pxcacheHash = NULL;

proxycacheItem *pxcache;
static int gline_fd = -1;
pthread_mutex_t mutex_pxcache;

extern pthread_mutex_t mutex_link;

void
proxycache_init()
{
   /* Init mutex */
   pthread_mutex_init(&mutex_pxcache, NULL);
   
   Debug((DBLMAX, "proxycache_init"));
   
   proxycache_load();
}

void
proxycache_free()
{
   if (pxcacheHash) {
      pthread_mutex_destroy(&mutex_pxcache);
      jenkins_hdestroy(pxcacheHash);
      pxcacheHash = NULL;
   }
}

void
proxycache_load()
{
   gheader_s gheader;
   proxycacheItem gitem;
   
   /* Create hash table */
   pxcacheHash = jenkins_hcreate(2);
   
   gline_fd = open("data/glines", O_RDONLY, 0);
   
   /* Try to read glines file header */
   if ((gline_fd < 0)
      || (read(gline_fd, &gheader, sizeof(gheader_s)) < sizeof(gheader_s))) {
      close(gline_fd);
      
      Debug((DBLMAX, "proxycache_init: default header"));
      
      gheader.total = 0;
      
      proxycache_openlogfile();
      write(gline_fd, &gheader, sizeof(gheader_s));
      
   } else {
      int i;
      
      Debug((DBLMAX, "proxycache_init: loading record file"));
      
      /* Load file... */
      for (i = 0; i < gheader.total; i++) {
         if (read(gline_fd, &gitem, sizeof(proxycacheItem)) < sizeof(proxycacheItem)) {
            fprintf(stderr, "proxycache_init: warning: glines file is probably corrupted\n");
            gheader.total = i;
            break;
         } else
            proxycache_insert(&gitem.addr, gitem.type, gitem.port, gitem.expire);
      }
      
      close(gline_fd);
      /* Now we open the log file for appending */
      proxycache_openlogfile();
   }
   
   /* Log file is opened for appending */
}

void
proxycache_openlogfile()
{
   gline_fd = open("data/glines", O_WRONLY | O_CREAT | O_APPEND, S_IREAD | S_IWRITE);
   if (gline_fd < 0) {
      fprintf(stderr, "proxycache_openlogfile: cannot open glines record file for writing\n");
      exit(1);
   }
}


void
proxycache_close()
{
   close(gline_fd);
}

/*
 * Write a gline record in the current log file
 */
void
proxycache_write(proxycacheItem *itemp)
{
   /* assert(gline_fd > 0); */
   write(gline_fd, itemp, sizeof(proxycacheItem));
}

/*
 * Save current glines in the glines record file
 */
void
proxycache_syncfile()
{
   time_t now;
   gheader_s gheader;
   proxycacheItem *citem;
   
   pthread_mutex_lock(&mutex_pxcache);
   Debug((DBLMAX, "proxycache_syncfile"));
   
   log_error("proxycache_syncfile: scanning pxcache");
   
   time(&now);
   gheader.total = 0;
   
   if (hfirst(pxcacheHash)) do {
      citem = hstuff(pxcacheHash);
      if (citem->expire > now)
         gheader.total++;
   } while (hnext(pxcacheHash));
   
   close(gline_fd);
   gline_fd = open("data/glines", O_WRONLY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
   
   /* Header */
   write(gline_fd, &gheader, sizeof(gheader_s));
   
   /* Data */
   if (hfirst(pxcacheHash)) do {
      citem = hstuff(pxcacheHash);
      if (citem->expire > now)
         proxycache_write(citem);
      else {
      /* Nice bug without cache_remove here: the proxy could come back after the
       * gline expired!
       */
         pthread_mutex_lock(&mutex_link);     /* Enter critical section */
         cache_remove(&citem->addr);          /* Recheck it baby! */
         pthread_mutex_unlock(&mutex_link);   /* Exit critical section */
      }
   } while (hnext(pxcacheHash));
   
   proxycache_close();
   
   /* Clear */
   if (hfirst(pxcacheHash)) do {
      xfree(hstuff(pxcacheHash));
   } while (hnext(pxcacheHash));
   
   jenkins_hdestroy(pxcacheHash);
   
   log_error("proxycache_syncfile: reloading");
   
   /* Reload */
   proxycache_load();
   
   log_error("proxycache_syncfile: exiting");
   pthread_mutex_unlock(&mutex_pxcache);
}



/* proxycache_insert()
 * Insert a new IP with type and expiration date info into the hash table
 * We use malloc here, and not a pre-allocation, but this is not annoying,
 * this function is "only" called when a proxy is found.
 */
void
proxycache_insert(
   struct in_addr *addr, /* IP */
   unsigned int type,    /* Type of proxy found */
   unsigned int port,    /* Port of http proxy, if any */
   time_t expire)        /* Absolute expiration date */
{
   
   proxycacheItem *citem;
   
   /* New element */
   citem = (proxycacheItem *)xmalloc(sizeof(proxycacheItem));
   
   /* Copy */
   memcpy(&citem->addr, addr, sizeof(struct in_addr));
   citem->type = type;
   citem->port = port;
   citem->expire = expire;
   
   /* Insert into hash table */
   if (hfind(pxcacheHash, addr, sizeof(struct in_addr))) {
      ((proxycacheItem *)hstuff(pxcacheHash))->expire = expire;
      xfree(citem); /* already added? */
      return;
   }
   
   if (!hadd(pxcacheHash, &citem->addr, sizeof(struct in_addr), citem))
      xfree(citem); /* already added? */
}

void
proxycache_insert_new(
   struct in_addr *addr, /* IP */
   unsigned int type,    /* Type of proxy found */
   unsigned int port,    /* Port of http proxy, if any */
   time_t expire)        /* Absolute expiration date */
{
   
   proxycacheItem *citem;
   
   /* New element */
   citem = (proxycacheItem *)xmalloc(sizeof(proxycacheItem));
   
   /* Copy */
   memcpy(&citem->addr, addr, sizeof(struct in_addr));
   citem->type = type;
   citem->port = port;
   citem->expire = expire;
   
   pthread_mutex_lock(&mutex_pxcache);
   
   /* Insert into hash table */
   if (!hadd(pxcacheHash, &citem->addr, sizeof(struct in_addr), citem))
      xfree(citem); /* already added? */
   else
      proxycache_write(citem);
   
   pthread_mutex_unlock(&mutex_pxcache);
}



/* Returns: 1 = success, 0 = failed
 */
int
proxycache_remove(struct in_addr *addr)
{
   int ret = 0;
   
   pthread_mutex_lock(&mutex_pxcache);
   
   if (hfind(pxcacheHash, addr, sizeof(struct in_addr))) {
      xfree(hstuff(pxcacheHash));
      ret = hdel(pxcacheHash);
   } else {
     log_error("proxycache_remove> addr not found");
   }
   
   pthread_mutex_unlock(&mutex_pxcache);
   return ret;
}


/* proxycache_check()
 * Need to be fast, but use hfind so it's ok. Called from the irc server thread
 * every time a new user connect, to see if it's a recorded insecure proxy (gline)
 * or not (nothing).
 *
 * -mbuna
 */
int
proxycache_check(struct in_addr *addr, proxycacheItem **ocitem)
{
   int ret = 0; /* default returns: recheck */
   
   if (hfind(pxcacheHash, addr, sizeof(struct in_addr))) {
      time_t now;
      proxycacheItem *citem = (proxycacheItem *)hstuff(pxcacheHash);
      time(&now);
      if (citem->expire > now) {
         *ocitem = citem;
         ret = 1; /* dont recheck */
      } else {
         xfree(hstuff(pxcacheHash));
         hdel(pxcacheHash); /* expired */
         cache_remove(addr); /* recheck - Thread issue: currently protected by mutex_link,
                              * so it's ok to use cache_remove here.
                              */
      }
   }
   return ret;
}




/* m_proxycache.c */
