/* Pxys - An IRC ProxyScan, m_scanner.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * Scanner node.
 *
 * $Id: m_scanner.c,v 1.24 2002/07/22 16:23:37 mbuna Exp $
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "conf.h"
#include "debug.h"

#include "typedefs.h"

#include "m_array.h"
#include "m_log.h"
#include "m_noscan.h"
#include "m_scanner.h"
#include "m_share.h"
#include "m_stats.h"
#include "m_thread.h"

#include "ms_globals.h"
#include "ms_msgout.h"       /* to inform the console channel */
#include "ms_netutils.h"

#include "mp_proxy.h"
#include "mp_socks.h"
#include "mp_wingate.h"



#define SET_CHECKING_STATE(x, y) x |= SCANTHREAD_CHECKING_##y
#define UNSET_CHECKING_STATE(x, y) x &= ~SCANTHREAD_CHECKING_##y;

extern pthread_mutex_t mutex_init;
extern pthread_mutex_t mutex_scanext;

long totalscanned = 0;
long totalfound = 0;
long totalcached = 0;

extern scanstats_s *scanstats;
extern time_t startscan_time;
time_t lastnojob = 0;
Boolean syncsent = false;


static void scanner_notify_netcheck();


void
scanner_startscan()
{
   scanthread_s *th;
   pthread_mutex_t mutex_dummy;
   
   pthread_mutex_init(&mutex_dummy, NULL);
   pthread_mutex_lock(&mutex_dummy) ;
   
   Debug((DBLBASIC, "scanner_startscan> %lu", (unsigned long) pthread_self()));
   
   pthread_mutex_lock(&mutex_init);
   th = get_scanthread(pthread_self());
   pthread_mutex_unlock(&mutex_init);
   
   if (NULL == th) {
      log_error("scanner_startscan: th == NULL !");
      exit(EXIT_FAILURE);
   }
   
   pthread_cond_wait(&th->pollSem, &mutex_dummy);
   pthread_cond_destroy(&th->pollSem);
   
   
   /* Endless loop... */
   for(;;) {
#ifdef SCAN_HTTP
      int i;
#endif
      
      pthread_mutex_lock(&mutex_scanext);
      array_fetchitem(th);
      pthread_mutex_unlock(&mutex_scanext); /* Allow others threads to get a next IP */
      
      if (th->counter == 0) {
         
         if (scanstats->waiting_threads >= 1) {
            pthread_mutex_lock(&mutex_init);  /* test&set */
            if (!syncsent) {
               syncsent = true;
               pthread_mutex_unlock(&mutex_init);
               scanner_notify_netcheck();
            } else
               pthread_mutex_unlock(&mutex_init);
         }
         
         /* Wait 2 seconds */
         sleep(2);
         
      } else {
        
         Debug((DBLMAX, "scanner_startscan: OK scan..."));
         
#ifdef SCAN_WINGATE
         SET_CHECKING_STATE(th->state, WINGATE);
         scan_wingate(th);
         UNSET_CHECKING_STATE(th->state, WINGATE);
#endif
#ifdef SCAN_SOCKS
         SET_CHECKING_STATE(th->state, SOCKS);
         scan_socks(th);
         UNSET_CHECKING_STATE(th->state, SOCKS);
#endif
#ifdef SCAN_HTTP
         for (i = 0; i < MyPerform->http_ports_count; i++) {
            th->port = MyPerform->http_ports[i];
            SET_CHECKING_STATE(th->state, HTTP);
            scan_proxy(th);
            UNSET_CHECKING_STATE(th->state, HTTP);
         }
#endif
         totalscanned += th->counter;
      }
   }
}

static void
scanner_notify_netcheck()
{
   time_t now;
   char locbuf[0x80];
   
   time(&now);
   human_time(now - startscan_time, locbuf);
   sendto_console("[%s] Network check completed. (%s)", Me->id, locbuf);

}


/* m_scanner.c */



