/* IRC ProxyScan, m_stats.h
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * $Id: m_stats.c,v 1.14 2002/07/22 16:23:37 mbuna Exp $
 */


#define GSTtemplate1 	"%ld week%s, %ld day%s, %ld hour%s"
#define GSTtemplate1b 	"%ld week%s, %ld day%s"
#define GSTtemplate1c 	"%ld week%s"
#define GSTtemplate2 	"%ld day%s, %ld hour%s, %ld min%s"
#define GSTtemplate2b 	"%ld day%s, %ld hour%s"
#define GSTtemplate2c 	"%ld day%s"
#define GSTtemplate3 	"%ld hour%s, %ld min%s"
#define GSTtemplate3b 	"%ld hour%s"
#define GSTtemplate4 	"%ld min%s, %ld sec%s"
#define GSTtemplate4b	"%ld min%s"
#define GSTtemplate5 	"%ld sec%s"

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "config.h"
#include "hashtab.h"
#include "m_stats.h"
#include "ms_myclient.h"
#include "ms_msgout.h"
#include "ms_globals.h"
#include "ms_netutils.h"
#include "ms_network.h"
#include "m_array.h"
#include "m_log.h"
#include "m_cache.h"
#include "m_proxycache.h"
#include "m_structs.h"
#include "misc.h"
#include "conf.h"
#include "typedefs.h"
/* xmalloc at the bottom of include files */
#include "xmalloc.h"

extern struct Client * mClient;
extern struct Server * mServer;
extern unsigned long clients_counter;
extern long ipa_current;
extern long totalscanned;
extern long totalfound;
extern long totalcached;
extern time_t startscan_time;
extern long timer_stats;
extern int msg_bogus;
extern pthread_mutex_t mutex_pxcache;
extern htab *pxcacheHash;
extern long serv_current;

extern server_handle_t mSA;
extern scanstats_s *scanstats;


struct statsrec_s {
   time_t offset_date;              /* offset date */
   long offset_scans;               /* number of scans */
   long offset_found;               /* number of (NEW) proxies found */
   long offset_cached;              /* number of cached proxy glined */
   unsigned long offset_clients;    /* number of new clients joining Undernet */
   double activity;                 /* activity percentile atm */
};

typedef struct statsrec_s 	statsrec_s;

/* Globals */
statistics_s *pxys_stats = NULL;
statsrec_s *statsrec = NULL;
unsigned short statsrec_pos;

void
stats_init()
{
   /* 288*300=86400
    * 288 records, with one record every 5 min (300s)
    * -> 24h stats records array
    */
   statsrec = (statsrec_s *)xcalloc(288, sizeof(statsrec_s));
   statsrec_pos = 0;
   
   pxys_stats = (statistics_s *)xmalloc(sizeof(statistics_s));
   memset(pxys_stats, 0, sizeof(statistics_s));
}

void
stats_free()
{
   if (statsrec) {
      xfree(statsrec);
      statsrec = NULL;
   }
   if (pxys_stats) {
      xfree(pxys_stats);
      pxys_stats = NULL;
   }
}

/* How the stats engine work:
 * f is the statistical func defined by statsrec[].
 * It stores the absolute numbers into f(pos);
 * when 5 min elapsed, these values are calculated to give
 * relative values (with the current one, stored in f(pos+1)), etc.
 *
 * Legal statsrec_pos : 0->287
 * Reserved statsrec_pos : 288 => statsrec[] full, perform translation.
 *
 * Array not full (statsrec_pos = i)
 * ---------------------------------------------------
 * | 0 | 1 | ... | i-2 | i-1 | ... | ... | n-2 | n-1 |
 * ---------------------------------------------------
 *   ...valid stats...    |   ...... all NULL ......
 *                        |
 *                similar to n-1 below
 *
 * Array in full mode:
 * ---------------------------------
 * | 0 | 1 | ... | ... | n-2 | n-1 |
 * ---------------------------------
 *  ..all are valid stats..    `this is the last absolute record, use
 *                              this ->date as the absolute date of the array
 */
void
stats_periodical_record(time_t now)
{
   double activity;
   long pxycached;
   
   activity = 100.0 * (double) scanstats->running_threads
            / (double) MyPerform->scanthreads;
   
   if (statsrec_pos > 0) {
      /* Calc relative values */
      statsrec_pos--;
      statsrec[statsrec_pos].offset_date =
           now - statsrec[statsrec_pos].offset_date;
      statsrec[statsrec_pos].offset_scans = 
           totalscanned - statsrec[statsrec_pos].offset_scans;
      statsrec[statsrec_pos].offset_found = 
           totalfound - statsrec[statsrec_pos].offset_found;
      statsrec[statsrec_pos].offset_cached = 
           totalcached - statsrec[statsrec_pos].offset_cached;
      statsrec[statsrec_pos].offset_clients = 
           clients_counter - statsrec[statsrec_pos].offset_clients;
      statsrec[statsrec_pos].activity =
           (statsrec[statsrec_pos].activity + activity)/2.0;
      
       pthread_mutex_lock(&mutex_pxcache);
      pxycached = hcount(pxcacheHash);
      pthread_mutex_unlock(&mutex_pxcache);
      
      /* scans:found:reissued_cached_gline:offsetclient:clients:servers:proxy_cache_size:activity */
      log_write(LOGID_STATS, "%ld:%ld:%ld:%ld:%ld:%ld:%ld:%g",
           statsrec[statsrec_pos].offset_scans,
           statsrec[statsrec_pos].offset_found,
           statsrec[statsrec_pos].offset_cached,
           statsrec[statsrec_pos].offset_clients,
           ipa_current,
           serv_current,
           pxycached,
           statsrec[statsrec_pos].activity);
      
      
      
      statsrec_pos++;
      /* Calc relative values -- fall through */
   }
   
   if (statsrec_pos == 288) {
      /* array is full, not a problem */
      memmove(&statsrec[0], &statsrec[1], 287 * sizeof(statsrec_s));
      statsrec_pos--;
   }
   
   statsrec[statsrec_pos].offset_date = now;
   statsrec[statsrec_pos].offset_scans = totalscanned;
   statsrec[statsrec_pos].offset_found = totalfound;
   statsrec[statsrec_pos].offset_cached = totalcached;
   statsrec[statsrec_pos].offset_clients = clients_counter;
   statsrec[statsrec_pos].activity = activity;
   
   if (statsrec_pos < 288)
      statsrec_pos++;
}


void
stats_dump(const char *nn)
{
   int i = 0;
   
   client_sendto(nn, "\2Dumping stats...\2");
   
   if (statsrec_pos > 50)
      i = statsrec_pos - 50;
   
   for ( ; (i < statsrec_pos) && statsrec[i].offset_date; i++)
      client_sendto(nn,
           "* offset:\2%d\2 scans:\2%ld\2 found:\2%ld\2 new clients:\2%ld\2 activity:\2%g%%\2",
           statsrec[i].offset_date, statsrec[i].offset_scans, statsrec[i].offset_found,
           statsrec[i].offset_clients, statsrec[i].activity);
}


void
consolestats()
{
   static long last_totalfound = 0;
   
   if (!MyConsole->channel)
      return;
   
   if (last_totalfound < totalfound) {
      last_totalfound = totalfound;
      sendto_console("[%s] [Clients: %ld] [Scans: %ld] [Found: %ld] [Cached: %ld]",
            Me->id, ipa_current + 1, totalscanned, totalfound, cache_status());
   }
}


/*
void
consolestats()
{
   long th_running, th_waiting;
   float thread_currentload;
   float foundpercent;
   
   if(NULL == MyConsole->channel)
      return;
   
   threadstats(&th_running, &th_waiting);
   thread_currentload = (100. * th_running) / (float) MyPerform->scanthreads;
   thread_avgload = (float) (((float) thread_avgload_count * thread_avgload)
                    + (float) thread_currentload) / (thread_avgload_count + 1);
   if(thread_avgload_count < 59)
      ++thread_avgload_count;
   
   foundpercent = (float) totalfound / (float) totalscanned;
   
   if(timer_stats >= CONSOLESTATS_DELAY) {
      timer_stats = 0;
      sendto_console("[%s] [Clients: %ld] [Scans: %ld] [Found: %ld (%2.0f%%)] [Load:%3.0f%% Avg:%3.0f%%]",
         Me->id, ipa_current, totalscanned, totalfound, foundpercent, thread_currentload, thread_avgload);
   }
}
*/

/* shit! */
void
human_time(time_t intime, char *outtime)
{
   time_t secs = 0;
   time_t mins = 0;
   time_t hrs = 0;
   time_t days = 0;
   time_t weeks = 0;
   
   *outtime = 0;
   
   if (intime >= 604800) {
		hrs = (intime / 3600) % 24 ;
		days = (intime / 86400 ) % 7 ;
		weeks = (intime / 604800 ) ;
		if(days == 0) {
			sprintf(outtime, GSTtemplate1c,
				weeks, weeks >= 2 ? "s" : "");
		} else if(hrs == 0) {
			sprintf(outtime, GSTtemplate1b,
				weeks, weeks >= 2 ? "s" : "",
				days, days >= 2 ? "s" : "");
		} else {
			sprintf(outtime, GSTtemplate1,
				weeks, weeks >= 2 ? "s" : "",
				days, days >= 2 ? "s" : "",
				hrs, hrs >= 2 ? "s" : "");
		}
	} else if(intime >= 86400) {
		mins = (intime / 60) % 60 ;
		hrs = (intime / 3600) % 24 ;
		days = (intime / 86400 ) ;
		if(hrs == 0) {
			sprintf(outtime, GSTtemplate2c,
				days, days >= 2 ? "s" : "");
		} else if(mins == 0) {
			sprintf(outtime, GSTtemplate2b,
				days, days >= 2 ? "s" : "",
				hrs, hrs >= 2 ? "s" : "");
		} else {
			sprintf(outtime, GSTtemplate2,
				days, days >= 2 ? "s" : "",
				hrs, hrs >= 2 ? "s" : "",
				mins, mins >= 2 ? "s" : "") ;
		}
	} else if(intime >= 3600) {
		mins = (intime / 60) % 60 ;
		hrs = (intime / 3600) % 24 ;
		if(mins == 0) {
			sprintf(outtime, GSTtemplate3b,
				hrs, hrs >= 2 ? "s" : "");
		} else {
			sprintf(outtime, GSTtemplate3,
				hrs, hrs >= 2 ? "s" : "",
				mins, mins >= 2 ? "s" : "");
		}
	} else if(intime >= 60) {
		secs = intime % 60 ;
		mins = (intime / 60) % 60 ;
		if(secs == 0) {
			sprintf(outtime, GSTtemplate4b,
				mins, mins >= 2 ? "s" : "");
		} else {
			sprintf(outtime, GSTtemplate4,
				mins, mins >= 2 ? "s" : "",
				secs, secs >= 2 ? "s" : "");
		}
	} else {
		sprintf(outtime, GSTtemplate5,
			intime, intime >= 2 ? "s" : "");
	}
}

#ifdef SCAN_HTTP
static void
stats_httpcount_build()
{
   if (pxys_stats->total_http == NULL) {
      if (MyPerform->http_ports_count > 0) {
         int i;
         pxys_stats->total_http = (struct portcount_s *)xmalloc(
               (MyPerform->http_ports_count + 1) * sizeof(struct portcount_s));
         for (i = 0; i < MyPerform->http_ports_count; i++) {
            pxys_stats->total_http[i].total = 0;
            pxys_stats->total_http[i].port = MyPerform->http_ports[i];
         }
         pxys_stats->total_http[i].total = 0;
         pxys_stats->total_http[i].port = 0;
      }
   }
}

void
stats_httpcount_addone(unsigned int port)
{
   int i;
   int found = 0;
   stats_httpcount_build();
   for (i = 0; pxys_stats->total_http[i].port > 0; i++) {
      if (pxys_stats->total_http[i].port == port) {
         ++(pxys_stats->total_http[i].total);
         found = 1;
         break;
      }
   }
   
   if (!found && pxys_stats->total_http) {
      ++i;
      pxys_stats->total_http = (struct portcount_s *)xrealloc(pxys_stats->total_http,
                                              (i + 1) * sizeof(struct portcount_s));
      pxys_stats->total_http[i-1].total = 1;
      pxys_stats->total_http[i-1].port = port;
      pxys_stats->total_http[i].total = 0;
      pxys_stats->total_http[i].port = 0;
      log_error("stats_httpcount_addone> added new port %u to http ports list", port);
   }
}
#endif


/* m_stats.c */
