/* Pxys - An IRC ProxyScan, mp_proxy.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * 
 * Module checking proxy/webserver tunelling vulnerability.
 *
 * $Id: mp_proxy.c,v 1.16 2002/07/19 13:34:20 mbuna Exp $
 */

#define HTTP_CONNECT_REQUEST "CONNECT %s:%d/ HTTP/1.1\nHost:%s:%d\n\n"

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "typedefs.h"
#include "conf.h"
#include "mp_proxy.h"
#include "mp_listener.h"
#include "mp_utils.h"
#include "ms_globals.h"
#include "ms_msgout.h"
#include "m_log.h"
#include "m_thread.h"
#include "debug.h"
#include "conf.h"
#include "mp_poll.h"   /* poll macros */

#include <errno.h>

extern int SPort;

extern pthread_mutex_t mutex_scanext;

extern int unreach_count;

void
scan_proxy(scanthread_s * sth)
{
   int err, nfds, i;
   struct pollfd *pfd;
   char request[96];
   struct timespec timewait;
   struct timespec timed;
   pthread_mutex_t pollMutex, pxyMutex;
   time_t now;
   unsigned int length;
   Boolean oneinprogress = false;
   Boolean stilloneinprogress = false;
   Boolean timed_out;
   
   /* - Mutex initialization */
   pthread_mutex_init(&pxyMutex, NULL);
   pthread_mutex_lock(&pxyMutex);
   
   /* - Building http proxy CONNECT request */
   if (Me->listeneraddr == htonl(INADDR_ANY))
      snprintf(request, sizeof(request), HTTP_CONNECT_REQUEST,
              Me->hostip, SPort, Me->hostip, SPort);
   else
      snprintf(request, sizeof(request), HTTP_CONNECT_REQUEST,
              Me->listener, SPort, Me->hostip, SPort);
           
   scanItems_init(sth, POLLWRITEFLAGS, (unsigned short)sth->port);
   
   INIT_SEMAPHOREWAITER(pollMutex, sth->pollSem);
   
   for (;;) {
#ifdef HAVE_GETNANOTIME
      getnanotime(&timed);
      timed.tv_sec++;
#else
      time(&timed.tv_sec);
      timed.tv_sec++;
      timed.tv_nsec = 0;
#endif
      pthread_cond_timedwait(&sth->pollSem, &pollMutex, &timed); 
      nfds = poll(sth->poll_fds, sth->counter, POLL_TIMEOUT);
      
      if (nfds <= -1) {
         if (errno == EINTR)
            continue;
         log_error("scan_proxy: poll error %d\n", errno);
         break;
      }
      
      /* Timeout checking */
      time(&now);
      timed_out = true;
      for (i = 0; i < sth->counter; ++i) {
         struct scanItem *sci = &sth->item[i];
         
         if (sci->state & (SIT_ERROR|SIT_FOUND|SIT_INPROGRESS))
            continue;
         
         if (now >= sth->item[i].timeout)
            sci->state |= SIT_TIMEOUT;
         
         timed_out = timed_out && (sci->state & SIT_TIMEOUT);
      }
      
      /* All fds are timed out, exiting. */
      if (timed_out)
         break;
      
      pfd = sth->poll_fds;
      
      for (i = 0; nfds && (i < sth->counter); ++i, ++pfd) {
         struct scanItem *sci = &sth->item[i];
         
         if (pfd->revents & POLLERRORS) {
            log_error("scan_proxy: POLLERRORS %s", inet_ntoa(sci->addr));
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if (sci->state & (SIT_ERROR|SIT_FOUND|SIT_INPROGRESS))
            continue;
         
         if ((err = get_sockerr(pfd->fd))) {
            if (err == EHOSTUNREACH) {
               unreach_count++;
         	   sci->state |= SIT_UNREACH;
         	   log_write(LOGID_UNREACH, "%s", inet_ntoa(sci->addr));
         	} else if (err != ECONNREFUSED) {
               log_error("scan_proxy: get_sockerr %s [%d] (%s)",
                    inet_ntoa(sci->addr), err, strerror(err));
            }
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if ((pfd->events & POLLWRITEFLAGS) && (pfd->revents & POLLWRITEFLAGS)) {
            /* Two step socket write */
            if (IsConnecting(sci)) {    /* Are we really connected? */
               SetConnected(sci);
               pfd->events = POLLWRITEFLAGS; /* We want to write */
               pfd->revents = 0;
               continue;
            }
            
            /* Write should not block or fail... let's go. */
            
            err = send_nonb(pfd->fd, request, strlen(request) + 1, &length);
            if (err < 0) {
               Debug((DBLMAX, "scan_proxy: send_nonb failed"));
               sci->state |= SIT_ERROR;
               DISABLE_PFD(pfd);
               nfds--;
               continue;
            } else if (err == 0) {
               pfd->events = POLLWRITEFLAGS; /* We want to try again */
               pfd->revents = 0;
               continue;
            }
            
            /* Set a new timeout delay. */
            time(&sci->timeout);
            sci->timeout += MyPerform->timeout_send;
            
            sci->state |= SIT_INPROGRESS;
            oneinprogress = true; /* to avoid items check (faster if nothing sent) */
            pfd->events = 0;
            DISABLE_PFD(pfd);
            nfds--;
         }
      }
   }
   
   /* If we were able to send at least one request to the http proxy server,
    * this thread has to wait for the reply (listener will notify us).
    */
   if (oneinprogress) {
      stilloneinprogress = false;
      for (i = 0; i < sth->counter; ++i) {
         if ((sth->item[i].state & SIT_INPROGRESS)
            && !(sth->item[i].state & SIT_RECEIVED)) {
            stilloneinprogress = true;
            break;
         }
      }
      
      if (stilloneinprogress) {
         pthread_cond_init(&sth->pxySem, NULL);  /* Create semaphore */
         time(&timewait.tv_sec);
         timewait.tv_sec += MyPerform->timeout_listen;
         timewait.tv_nsec = 0;
         
         sth->state |= SCANTHREAD_TIMEWAIT;
         
         do {
            err = pthread_cond_timedwait(&sth->pxySem, &pxyMutex, &timewait);
            if (err == ETIMEDOUT) {
               Debug((DBLMAX, "scan_proxy: timeout"));
               break;
            }
            
            Debug((DBLMAX, "scan_proxy: Reply received!!!"));
            
            stilloneinprogress = false;
            for (i = 0; i < sth->counter; ++i) {
               if ((sth->item[i].state & SIT_INPROGRESS)
                  && !(sth->item[i].state & SIT_RECEIVED)) {
                  stilloneinprogress = true;
                  break;
               }
            }
         } while (stilloneinprogress);
      
         sth->state &= ~SCANTHREAD_TIMEWAIT;
         pthread_cond_destroy(&sth->pxySem);
      }
   }
   scanItems_closeall(sth);
   DESTROY_SEMAPHOREWAITER(pollMutex, sth->pollSem);
   pthread_mutex_unlock(&pxyMutex);
   pthread_mutex_destroy(&pxyMutex); /* no memory leak plz k thx. */
}


/* m_proxy.c */
