/* Pxys - An IRC ProxyScan, mp_socks.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * E-mail of the author: mbuna@undernet.org
 *
 *
 * Unsecure socks servers version 4 and 5 scanner.
 *
 * $Id: mp_socks.c,v 1.25 2002/04/15 13:09:43 mbuna Exp $
 */


#define SOCKS_PORT      1080

/* Removed TOAST_REJECTED_SOCKS4, now, I'm glining every replyed socks replied
 * as socks v4 with an status code != rejected. It should avoid tadd's gline :P~
 */


#include "config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <errno.h>

#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "conf.h"
#include "debug.h"

#include "typedefs.h"

#include "mp_socks.h"
#include "mp_utils.h"
#include "mp_poll.h"

#include "m_log.h"
#include "m_thread.h"

#include "ms_globals.h"
#include "ms_msgout.h"

#include "socks4.h"             /* Socks4 protocol constants */
#include "socks5.h"             /* Socks5 protocol constants */

extern pthread_mutex_t mutex_scanext;

extern scanstats_s *scanstats;

extern int unreach_count;

/* Why we cannot check socks version 4 then version5 in one connection:

The SOCKS server closes its connection immediately after notifying
the client of a failed or rejected request. For a successful request,
the SOCKS server gets ready to relay traffic on both directions. This
enables the client to do I/O on its connection as if it were directly
connected to the application server.

http://www.socks.nec.com/protocol/socks4.protocol

->04 01 1A 0B C2 75 C2 78 00 
<-04 5A 1A 0B C2 75 C2 78 
SOCKS firewall: connection to ircu.clubint.net granted.

*/

void
scan_socks(scanthread_s *sth)
{
   int nfds, i, err;
   struct pollfd *pfd;
   unsigned int length;
   struct timespec timed;
   pthread_mutex_t pollMutex;
   time_t now;
   int timed_out;
   
   /* Socks4 data init */
   struct socks4 socks4out = { S4_VN, S4_CMD_CONNECT, { 0x1A, 0x0B }, { 0, 0, 0, 0 }, 0 };
   struct socks4_response socks4in = { 0, 0, { 0, 0 }, { 0, 0, 0, 0 } };
   
   /* Socks5 data init */
   struct socks5 socks5out = { S5_VN, S5_CMD_CONNECT };
   struct socks5_response socks5in = { 0, 0 };
   
   /* To set the DST IP of the socks4 CONNECT request, we choose in the MyTarget list.
    * ie. if for Undernet we put an Undernet servers list, we are 100% sure that if
    * the request is granted by the SOCKS 4, it's misconfigured.
    * Without an Undernet server IP for the Undernet network, it would only be speculation.
    *
    * One day after: OK, this code (get_socks4_dest_ip) is in fact unuseful. SOCKS4 will
    * often reject the connection request, see the note for TOAST_REJECTED_SOCKS4 (at
    * the top of this page). So, even if rejected (but only if accepted OR rejected,
    * others reply are not considered as insecure SOCKS 4 proxy), we toast 'em.  --mbuna
    *
    * We let get_socks4_dest_ip for the fun (and because accepted request may hit a
    * specified server... with more servers, it's better).
    */
   
   get_socks4_dest_ip(&socks4out);
   
   /************************* SOCKS 4 CHECK ********************************/
   
   
   scanItems_init(sth, POLLWRITEFLAGS, SOCKS_PORT);
   
   INIT_SEMAPHOREWAITER(pollMutex, sth->pollSem);
   
   Debug((DBLMAX, "scan_socks (SOCKS 4) sscanItems_init ok"));
   
   for (;;) {
#ifdef HAVE_GETNANOTIME
      getnanotime(&timed);
      timed.tv_sec++;
#else
      time(&timed.tv_sec);
      timed.tv_sec++;
      timed.tv_nsec = 0;
#endif
      pthread_cond_timedwait(&sth->pollSem, &pollMutex, &timed);
      nfds = poll(sth->poll_fds, sth->counter, POLL_TIMEOUT);
      
      if (nfds < 0) {
         if (errno == EINTR)
            continue;
         log_error("scan_socks4: poll error (%d)", errno);
         break;
      }
      
      /* Timeout checking */
      time(&now);
      timed_out = true;
      for (i = 0; i < sth->counter; i++) {
         struct scanItem *sci = &sth->item[i];
         
         if (sci->state & (SIT_ERROR|SIT_FOUND)) /* skip inactives */
            continue;
         
         if (now >= sci->timeout)
            sci->state |= SIT_TIMEOUT;
         
         timed_out = timed_out && (sci->state & SIT_TIMEOUT);
      }
      
      /* All fds are timed out, exiting. */
      if (timed_out)
         break;
      
      pfd = sth->poll_fds;
      
      for (i = 0; nfds && (i < sth->counter); ++i, ++pfd) {
         struct scanItem *sci = &sth->item[i];
         
         if (pfd->revents & POLLERRORS) {
            log_error("mp_socks:socks4: POLLERRORS %s", inet_ntoa(sci->addr));
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if (sci->state & (SIT_ERROR|SIT_FOUND)) /* skip inactives */
            continue;
         
         if ((err = get_sockerr(pfd->fd))) {
            if (err == EHOSTUNREACH) {
               unreach_count++;
               sci->state |= SIT_UNREACH;
               log_write(LOGID_UNREACH, "%s", inet_ntoa(sci->addr));
            } else if (err != ECONNREFUSED) {
               log_error("mp_socks:socks4: get_sockerr %s [%d] (%s)",
                    inet_ntoa(sci->addr), err, strerror(err));
            }
            sci->state |= SIT_ERROR | SIT_NOSOCKS;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if ((pfd->events & POLLWRITEFLAGS) && (pfd->revents & POLLWRITEFLAGS)) {
            
            if (IsConnecting(sci)) {    /* Are we really connected? */
               SetConnected(sci);
               pfd->events = POLLWRITEFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            /* socket writable */
            
            Debug((DBLMAX, "->%s %X %X %X %X %X %X %X %X %X", inet_ntoa(sci->addr),
               socks4out.vn, socks4out.cd, socks4out.dstport[0], socks4out.dstport[1],
               socks4out.dstip[0], socks4out.dstip[1], socks4out.dstip[2], socks4out.dstip[3],
               socks4out.nullfield));
            
            err = send_nonb(pfd->fd, (char *) &socks4out, sizeof(struct socks4), &length);
            if (err < 0) {
               Debug((DBLMAX, "scan_socks4: %s send_nonb failed", inet_ntoa(sci->addr)));
               sci->state |= SIT_ERROR | SIT_NOSOCKS;
               DISABLE_PFD(pfd);
               nfds--;
               continue;
            } else if (err == 0) {
               /* We want to try again. */
               pfd->events = POLLWRITEFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            time(&sci->timeout);
            sci->timeout += MyPerform->timeout_send;
            
            /* Now we want to read the reply. */
            pfd->events = POLLREADFLAGS;
            pfd->revents = 0;
            continue;
         }
         
         if ((pfd->events & POLLREADFLAGS) && (pfd->revents & POLLREADFLAGS)) {
            memset((void *)&socks4in, 0, sizeof(struct socks4_response));
            err = recv_nonb(pfd->fd, (char *) &socks4in, sizeof(struct socks4_response), &length);
            if (err < 0) {
               Debug((DBLMAX, "scan_socks4: %s recv_nonb failed", inet_ntoa(sci->addr)));
               sci->state |= SIT_ERROR;
               DISABLE_PFD(pfd);
               nfds--;
               continue;
            } else if (err == 0) {
               /* We want to try again. */
               pfd->events = POLLREADFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            if (length > 1) {
               /* We are interested at the 2 first bytes. */
               
               /* Debugging... */
               Debug((DBLMAX,
               "scan_socks4: %s recv_nonb %d (VN=%d CD=%d dstport = %d %d dstip = %d.%d.%d.%d)",
                  inet_ntoa(sci->addr), length,
                  (int)socks4in.vn, (int)socks4in.cd,
                  (int)socks4in.dstport[0], (int)socks4in.dstport[1],
                  (int)socks4in.dstip[0], (int)socks4in.dstip[1],
                  (int)socks4in.dstip[2], (int)socks4in.dstip[3]));
               
               /* Valid socks4 reply with access granted.*/
               if (socks4in.cd == S4_REQ_GRANTED) {
                  /* Toast it. */
                  sci->state |= SIT_FOUND;
                  proxy_toast(&sci->addr, PROXY_TYPE_SOCKS_V4, 0);
                  
               } else {
                  /* Logging for development purposes */
                  log_error("socks4in: (rejected) %s %02X %02X %02X %02X %02X %02X %02X %02X",
                      inet_ntoa(sci->addr), socks4in.vn, socks4in.cd,
                      socks4in.dstport[0], socks4in.dstport[1],
                      socks4out.dstip[0], socks4out.dstip[1],
                      socks4out.dstip[2], socks4out.dstip[3]);
               }
            } else {
               /* We got a tinny reply.. that's not a socks server, we should be happy! */
               log_error("mp_socks:socks4 recv_nonb -> length=%d", length);
            }
            /* close connection immediately */
            DISABLE_PFD(pfd);
            nfds--;
         }
      }
   }
   
   scanItems_closeall(sth);
   
   DESTROY_SEMAPHOREWAITER(pollMutex, sth->pollSem);
   
   /************************* SOCKS 5 CHECK ********************************/
   
   scanItems_init(sth, POLLWRITEFLAGS, SOCKS_PORT);
   
   INIT_SEMAPHOREWAITER(pollMutex, sth->pollSem);
   
   for (;;) {
#ifdef HAVE_GETNANOTIME
      getnanotime(&timed);
      timed.tv_sec++;
#else
      time(&timed.tv_sec);
      timed.tv_sec++;
      timed.tv_nsec = 0;
#endif
      
      pthread_cond_timedwait(&sth->pollSem, &pollMutex, &timed);
      nfds = poll(sth->poll_fds, sth->counter, POLL_TIMEOUT);
      
      if (nfds <= -1) {
         if (errno == EINTR)
            continue;
         log_error("scan_socks5: poll error %d\n", errno);
         break;
      }
      
      /* Timeout checking */
      time(&now);
      timed_out = true;
      for (i = 0; i < sth->counter; ++i) {
         struct scanItem *sci = &sth->item[i];
         
         if (sci->state & (SIT_ERROR|SIT_FOUND)) /* skip inactives */
            continue;
         
         if (now >= sci->timeout)
            sci->state |= SIT_TIMEOUT;
         
         timed_out = timed_out && (sci->state & SIT_TIMEOUT);
      }
      
      if (timed_out)
         break;
      
      pfd = sth->poll_fds;
      
      for (i = 0; nfds && (i < sth->counter); ++i, ++pfd) {
         struct scanItem *sci = &sth->item[i];
         
         if (pfd->revents & POLLERRORS) {
            log_error("mp_socks:socks5: POLLERRORS %s", inet_ntoa(sci->addr));
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if (sci->state & (SIT_ERROR|SIT_FOUND)) /* skip inactives */
            continue;
         
         if ((err = get_sockerr(pfd->fd))) {
            if (err == EHOSTUNREACH) {
               unreach_count++;
               sci->state |= SIT_UNREACH;
               log_write(LOGID_UNREACH, "%s", inet_ntoa(sci->addr));
               log_error("mp_socks:socks5: %s EHOSTUNREACH", inet_ntoa(sci->addr));
            } else if (err != ECONNREFUSED) {
               log_error("mp_socks:socks5: get_sockerr %s [%d] (%s)",
                    inet_ntoa(sci->addr), err, strerror(err));
            }
            sci->state |= SIT_ERROR | SIT_NOSOCKS;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if ((pfd->events & POLLWRITEFLAGS) && (pfd->revents & POLLWRITEFLAGS)) {
            
            if (IsConnecting(sci)) {    /* Are we really connected? */
               SetConnected(sci);
               pfd->events = POLLWRITEFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            /* socket writable */
            
            err = send_nonb(pfd->fd, (char *) &socks5out, sizeof(struct socks5), &length);
            if (err < 0) {
               sci->state |= SIT_ERROR;
               DISABLE_PFD(pfd);
               nfds--;
               continue;
            } else if (err == 0) {
               pfd->events = POLLWRITEFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            time(&sci->timeout);
            sci->timeout += MyPerform->timeout_send;
            
            /* Now we want to read the reply. */
            pfd->events = POLLREADFLAGS;
            pfd->revents = 0;
            continue;
         }
         
         if ((pfd->events & POLLREADFLAGS) && (pfd->revents & POLLREADFLAGS)) {
            memset((void *)&socks5in, 0, sizeof(struct socks5_response));
            err = recv_nonb(pfd->fd, (char *) &socks5in, sizeof(struct socks5_response), &length);
            if (err < 0) {
               sci->state |= SIT_ERROR;
               DISABLE_PFD(pfd);
               nfds--;
               continue;
            } else if (err == 0) {
               pfd->events = POLLREADFLAGS;
               pfd->revents = 0;
               continue;
            }
            
            if (length > 1) {
               /* We are interested at the 2 first bytes. */
               if (socks5in.vn == S5_VN) {
                  switch(socks5in.method) {
                     case S5_R_METHOD_NOAUTH:
                        /* Toast it. */
                        sci->state |= SIT_FOUND;
                        proxy_toast(&sci->addr, PROXY_TYPE_SOCKS_V5, 0);
                        break;
                     case S5_R_METHOD_GSSAPI:
                        scanstats->socks5_gssapi_count++;
                        log_error("socks5in: (GSSAPI) %s %02X %02X",
                           inet_ntoa(sci->addr), socks5in.vn, socks5in.method);
                        break;
                     case S5_R_METHOD_USERPW:
                        scanstats->socks5_userpw_count++;
                        log_error("socks5in: (USER/PW) %s %02X %02X",
                           inet_ntoa(sci->addr), socks5in.vn, socks5in.method);
                        break;
                     case S5_R_METHOD_NOTACC:
                        scanstats->socks5_notacc_count++;
                        log_error("socks5in: (NOTACC) %s %02X %02X",
                           inet_ntoa(sci->addr), socks5in.vn, socks5in.method);
                        break;
                     default:
                        /* Other shits at port 1080 */
                        Debug((DBLMAX, "scan_socks:socks5: no valid type"));
                  }
               }
            } else {
               log_error("mp_socks:socks5 recv_nonb -> length=%d", length);
            }
            DISABLE_PFD(pfd);
            nfds--;
         }
      }
   }
   
   /* Remove SIT_NOSOCKS flag (used to not try to connect twice in socks check) */
   for (i = 0; i < sth->counter; ++i) {
      if (sth->item[i].state & SIT_NOSOCKS)
            sth->item[i].state &= ~SIT_NOSOCKS;
   }
   
   scanItems_closeall(sth);
   
   DESTROY_SEMAPHOREWAITER(pollMutex, sth->pollSem);
}


void
get_socks4_dest_ip(struct socks4 * outdata)
{
   int i, chooser = (int) ((double)totalTarget*rand()/(RAND_MAX+1.0));
   Target_s *target = MyTarget;
   
   for (i = chooser; target && i--; target = target->next)
      ;
   
   if (!target) /* should never happen */
      target = MyTarget;
   
   memcpy(&outdata->dstip, &target->addr, sizeof(outdata->dstip));
   Debug((DBLMAX, "mp_socks:get_socks4_dest_ip: %u.%u.%u.%u",
            outdata->dstip[0], outdata->dstip[1], outdata->dstip[2], outdata->dstip[3])); 
}

/* mp_socks.c */
