/* Pxys - An IRC ProxyScan, mp_wingate.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * Classic WinGate> scanner.
 *
 * $Id: mp_wingate.c,v 1.14 2002/04/15 12:39:51 mbuna Exp $
 */


#define WINGATE_PORT	23

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "conf.h"
#include "debug.h"
#include "typedefs.h"

#include "mp_poll.h"
#include "mp_wingate.h"
#include "mp_utils.h"
#include "m_log.h"
#include "m_shutdown.h"
#include "m_thread.h"
#include "ms_globals.h"
#include "ms_msgout.h"

extern int unreach_count;


void
scan_wingate(scanthread_s * sth)
{
   int i, nfds, err;
   unsigned int length;
   time_t now;
   struct pollfd* pfd;
   struct timespec timed;
   pthread_mutex_t pollMutex;
   char winreadbuf[0x80];
   Boolean result = false;
   Boolean timed_out;
   
   scanItems_init(sth, POLLWRITEFLAGS, WINGATE_PORT);
   
   INIT_SEMAPHOREWAITER(pollMutex, sth->pollSem);
    
   /* And start poll lookup loop */
   for (;;) {
#ifdef HAVE_GETNANOTIME
      getnanotime(&timed);
      timed.tv_sec++;
#else
      time(&timed.tv_sec);
      timed.tv_sec++;
      timed.tv_nsec = 0;
#endif
      pthread_cond_timedwait(&sth->pollSem, &pollMutex, &timed);
      nfds = poll(sth->poll_fds, sth->counter, POLL_TIMEOUT);
      
      if (nfds <= -1) {
         if (errno == EINTR)
            continue;
         log_error("scan_wingate: poll error %d (%s)\n", errno, strerror(errno));
         break;
      }
      
      /* Timeout checking */
      time(&now);
      timed_out = true;
      for (i = 0; i < sth->counter; i++) {
         struct scanItem *sci = &sth->item[i];
         
         if (sci->state & (SIT_ERROR|SIT_FOUND))
            continue;
         if (now >= sci->timeout)
            sci->state |= SIT_TIMEOUT;
         
         timed_out = timed_out && (sci->state & SIT_TIMEOUT);
      }
      
      /* All fds are timed out, exiting. */
      if (timed_out)
         break;
      
      pfd = sth->poll_fds;

      for (i = 0; nfds && (i < sth->counter); ++i, ++pfd) {
         struct scanItem *sci = &sth->item[i];
         
         if (pfd->revents & POLLERRORS) {
            log_error("mp_wingate: POLLERRORS %s", inet_ntoa(sci->addr));
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if (sci->state & (SIT_ERROR|SIT_FOUND)) /* skip inactives */
            continue;
         
         if ((err = get_sockerr(pfd->fd))) {
            if (err == EHOSTUNREACH) {
               unreach_count++;
               sci->state |= SIT_UNREACH;
               log_write(LOGID_UNREACH, "%s", inet_ntoa(sci->addr));
            } else if (err != ECONNREFUSED) {
               log_error("mp_wingate: get_sockerr %s [%d] (%s)",
                    inet_ntoa(sci->addr), err, strerror(err));
            }
            sci->state |= SIT_ERROR;
            DISABLE_PFD(pfd);
            nfds--;
            continue;
         }
         
         if ((pfd->events & POLLWRITEFLAGS) && (pfd->revents & POLLWRITEFLAGS)) {
            if (IsConnecting(sci)) {    /* Are we really connected? */
               SetConnected(sci);
               pfd->events = POLLREADFLAGS; /* We want to read */
               pfd->revents = 0;
               continue;
            }
         }
         
         if ((pfd->events & POLLREADFLAGS) && (pfd->revents & POLLREADFLAGS)) {
            
            /* 6- Read ready: receive data */
            memset(winreadbuf, 0, sizeof(winreadbuf));
            err = recv_nonb(pfd->fd, winreadbuf, sizeof(winreadbuf), &length);
            if (err > 0) {
               sci->tryread++;
               if (length > 0) {
                  if (!strncmp(winreadbuf, "WinGate>", 8) || !strncmp(winreadbuf, "Too m", 5)) {
                     sci->state |= SIT_FOUND;
                     proxy_toast(&sci->addr, PROXY_TYPE_WINGATE, 0);
                     Debug((DBLMAX, "scan_wingate:recv_nonb Wingate found (%d): %s",
                        sci->tryread, winreadbuf));
                     result = true;
                     sci->tryread = 2;
                  }
                  if (sci->tryread >= 2) {
                     shutdown(pfd->fd, SHUT_RDWR);
                     DISABLE_PFD(pfd);
                     nfds--;
                     continue;
                  }
               }
            } else if (err == 0) {
               /* We want to retry. */
               pfd->events = POLLREADFLAGS;
               pfd->revents = 0;
            } else {
               DISABLE_PFD(pfd);
               nfds--;
            }
         }
      }
   }
   scanItems_closeall(sth);
   DESTROY_SEMAPHOREWAITER(pollMutex, sth->pollSem);
}

/* mp_wingate.c */
