/* Pxys - An IRC ProxyScan, ms_mode.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * $Id: ms_mode.c,v 1.10 2002/04/21 15:22:29 mbuna Exp $
 */
 
#include "config.h"


#include "ms_mode.h"
#include "m_structs.h"
#include "ms_msgout.h"
#include "ms_netutils.h"
#include "m_array.h"
#include "m_log.h"

#include "hashtab.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

extern htab *mHC;
extern server_handle_t mSA;
extern pthread_mutex_t mutex_link;

inline void showopr(struct Client *oper);


/* ms_mode is always called in the same thread as addclient/delclient */
void
ms_mode(const char * ClientYYXXX, const char * mode)
{
   struct Client *cptr;
   
#ifdef N2K_SUPPORT
   cptr = array_findclient_from_numeric(ClientYYXXX);
#else
   if (*ClientYYXXX == ':')
      ClientYYXXX++;
   cptr = array_findclient_from_nick(ClientYYXXX);
#endif
   
   if (NULL == cptr) {
      log_error("ms_mode: user %s not found!\n", ClientYYXXX);
   } else {
      int oper;
      
      oper = (cptr->flags & FLAG_OPER) ? 1 : 0;
      
      operchange(&oper, mode);
      
      if (oper && !(cptr->flags & FLAG_OPER)) {
         cptr->flags |= FLAG_OPER;
         showopr(cptr);
      } else if (!oper && (cptr->flags & FLAG_OPER))
         cptr->flags &= ~FLAG_OPER;
   }
}


/* mode != NULL */
void
operchange(int * ioOper, const char * mode)
{
   int tog = 1;
   int ch;
   
   while((ch = *mode++))
      switch(ch) {
         case '+':
            tog = 1;
            continue;
         case '-':
            tog = 0;
            continue;
         case 'o':
            *ioOper = tog;
   }
}

/* */
inline void
showopr(struct Client *oper)
{
   pthread_mutex_lock(&mutex_link);
   if (hfirst(mHC)) do {
       struct Client *cptr = (struct Client *) hstuff(mHC);
	   if (cptr->flags & FLAG_SHOWOPR) {
	      char buf[6], *servername;
	      numnick(cptr, buf);
	      if (mSA[oper->nserv])
	         servername = ((struct Server *)mSA[oper->nserv])->name;
	      else
	         servername = "*DESYNC*";
	      
	      client_sendto(buf, "\2OPR\2 %s [%s] on %s", oper->nick,
	         inet_ntoa(oper->ip), servername);
	   }
   } while(hnext(mHC));
   pthread_mutex_unlock(&mutex_link);
}


/* ms_mode.c */
